# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import io
import abc
import six
import sys
import shutil
import tempfile

from .. import errors
from ..mode import Mode
from ..copy import copy_file
from ..path import abspath, normpath, join, dirname
from ..base import FS
from ..opener import open_fs
from ..wrapfs import WrapFS
from ..multifs import MultiFS
from .._fscompat import fsdecode, fspath

from .wrap import WrapWritable
from ._utils import writable_stream, writable_path, unique


@six.add_metaclass(abc.ABCMeta)
class ArchiveSaver(object):
    """Base class for archive serializers.
    """

    def __init__(self, output, overwrite=False, initial_position=0, **options):
        """Create a new serializer.

        Parameters:
            output (`io.IOBase` or `str`): the filename of the destination
                or a file handle in which to write the archive.
            overwrite (`boolean`): If True, use a temporary file to save
                the contents of the archive. Useful when updating an archive
                file. **[default: False]**
            initial_position (`int`): The initial position of the stream when
                it was seen for the first time. **[default: 0]**

        """
        self.output = output
        self.overwrite = overwrite
        self.initial_position = initial_position
        self.stream = isinstance(output, io.IOBase)

    def save(self, fs):
        """Save the given FS.

        Parameters:
            fs (`fs.base.FS`): the filesystem to save in the archive.

        """
        if self.stream:
            self.to_stream(fs)
        else:
            self.to_file(fs)

    def to_file(self, fs):
        """Save the given FS, considering ``self.output`` as a filename.

        Parameters:
            fs (`fs.base.FS`): the filesystem to save in the archive.

        """
        if self.overwrite: # If we need to overwrite, use temporary file
            tmp = '.'.join([self.output, 'tmp'])
            self._to(tmp, fs)
            shutil.move(tmp, self.output)
        else:
            self._to(self.output, fs)

    def to_stream(self, fs):
        """Save the given FS, considering ``self.output`` as a stream.

        Parameters:
            fs (`fs.base.FS`): the filesystem to save in the archive.

        """
        if self.overwrite: # If we need to overwrite, use temporary file
            fd, temp = tempfile.mkstemp()
            os.close(fd)
            self._to(temp, fs)

            self.output.seek(self.initial_position)
            with open(temp, 'rb') as f:
                shutil.copyfileobj(f, self.output)

            os.remove(temp)

        else:
            self._to(self.output, fs)

    @abc.abstractmethod
    def _to(self, handle, fs):
        """Save the given FS to the given stream handle.

        Parameters:
            handle (`io.IOBase`): a writable stream in which to write
                the filesystem in an archive.
            fs (`fs.base.FS`): the filesystem to save.

        """
        raise NotImplementedError()


@six.add_metaclass(abc.ABCMeta)
class ArchiveReadFS(FS):
    """A filesystem allowing to read an archive.
    """

    _meta = NotImplemented

    def __init__(self, handle, **options):
        """Create a new archive reader filesystem.

        Parameters:
            handle (`io.IOBase` or `str`): a filename or a readable
                file-like object storing the archive to read

        Keyword Arguments:
            close_handle (`boolean`): if ``True``, close the handle
                when the filesystem is closed. **[default: True]**

        """
        super(ArchiveReadFS, self).__init__()

        if isinstance(handle, six.binary_type):
            # Decode the path if it is in binary format
            handle = fsdecode(fspath(handle))

        if isinstance(handle, six.text_type):
            # Expand the path
            _path = os.path.expanduser(os.path.expandvars(handle))
            _path = os.path.normpath(os.path.abspath(_path))
            # Create the readable fs if the handle exists
            if os.path.exists(_path) and os.access(_path, os.R_OK):
                self._close_handle = True
                self._handle = open(_path, 'rb')

        elif hasattr(handle, 'read'):
            # Create the readable fs if the handle is readable
            if handle.readable() and handle.seekable():
                self._close_handle = options.get('close_handle', True)
                self._handle = handle

        else:
            raise errors.CreateFailed("cannot use {}".format(handle))

    def __repr__(self):  # noqa: D105
        return "{}({!r})".format(
            self.__class__.__name__,
            getattr(self._handle, 'name', self._handle),
        )

    def __str__(self):  # noqa: D105
        return "<{} '{}'>".format(
            self.__class__.__name__.lower(),
            getattr(self._handle, 'name', self._handle),
        )

    def _on_modification_attempt(self, path):
        """Call when a modification is attempted on the archive.
        """
        raise errors.ResourceReadOnly(path)

    def setinfo(self, path, info):  # noqa: D102
        self.check()
        self._on_modification_attempt(path)

    def makedir(self, path, permissions=None, recreate=False):  # noqa: D102
        self.check()
        self._on_modification_attempt(path)

    def remove(self, path):  # noqa: D102
        self.check()
        self._on_modification_attempt(path)

    def removedir(self, path):  # noqa: D102
        self.check()
        self._on_modification_attempt(path)

    def getmeta(self, namespace="standard"):  # noqa: D102
        if namespace == "standard":
            return self._meta['standard'].copy()
        return {}

    def close(self):  # noqa: D102
        if not self.isclosed():
            if self._close_handle:
                getattr(self._handle, 'close', lambda: None)()
            super(ArchiveReadFS, self).close()


@six.add_metaclass(abc.ABCMeta)
class ArchiveFS(WrapFS):
    """A wrapper filesystem allowing to read, write and update an archive.
    """
    
    _read_fs_cls = NotImplemented
    _saver_cls = NotImplemented

    def __init__(self, handle, proxy=None, **options):
        """Create a new archive filesystem.

        Parameters:
            handle (io.IOBase or str): a filename or a stream storing an
                archive and/or in which to write the updated archive.
            proxy (FS): the filesystem to use as to perform temporary
                write operations. Leave to `None` to use the default
                defined in `~fs.archive.wrap.WrapWritable`.
                **[default: `~fs.memoryfs.MemoryFS`]**

        Keyword Arguments:
            close_handle (boolean): if `True`, close the handle
                when the filesystem is closed. **[default: True]**

        """
        initial_position = 0
        read_fs = None
        self._saver = None

        if isinstance(handle, six.binary_type):
            # Decode the path if it is in binary format
            handle = fsdecode(fspath(handle))

        if isinstance(handle, six.text_type):
            # Expand the path
            _path = os.path.expanduser(os.path.expandvars(handle))
            _path = os.path.normpath(os.path.abspath(_path))
            # Create the readable fs if the handle exists
            if os.path.exists(_path) and os.access(_path, os.R_OK):
                options.setdefault('close_handle', True)
                read_fs = self._read_fs_cls(handle, **options)
            # Create a saver only if the destination is writable
            create_saver = writable_path(_path)

        elif hasattr(handle, 'read'):
            # Get the initial stream position
            initial_position = getattr(handle, 'tell', lambda: 0)()
            # Create the readable fs if the handle is readable
            if handle.readable() and handle.seekable():
                read_fs = self._read_fs_cls(handle, **options)
            # Create a saver only if the destination is writable
            create_saver = writable_stream(handle)

        else:
            raise errors.CreateFailed("cannot use {}".format(handle))

        overwrite = read_fs is not None
        if create_saver:
            self._saver = self._saver_cls(handle, overwrite, initial_position)

        proxy = proxy or "mem://"
        wrapped_fs = WrapWritable(read_fs, writable_fs=proxy) \
                  if read_fs is not None else open_fs(proxy)
        super(ArchiveFS, self).__init__(wrapped_fs)

    def close(self):  # noqa: D102
        if not self.isclosed():
            if self._saver is not None:
                self._saver.save(self)
            self.delegate_fs().close()
            super(ArchiveFS, self).close()



# coding: utf-8
"""Tar archive filesystems.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import sys
import time
import tarfile
import datetime

import six

from ... import errors
from ...info import Info
from ...mode import Mode
from ...time import datetime_to_epoch
from ...path import basename, relpath, splitext, isbase, parts, frombase
from ...enums import ResourceType
from ...permissions import Permissions

from .. import base
from .._utils import unique

from .iotools import RawWrapper
from .tarfile2 import TarFile


class TarReadFS(base.ArchiveReadFS):
    """A read-only filesystem within a TAR archive.
    """

    _meta = {
        'standard': {
            'case_insensitive': True, # FIXME : is it ?
            'network': False,
            'read_only': True,
            'supports_rename': False,
            'thread_safe': True, # FIXME: is it ?
            'unicode_paths': True,
            'virtual': False,
            'max_path_length': None,
            'max_sys_path_length': None,
            'invalid_path_chars': '\x00\x01',
        },
    }

    _TYPE_MAP = {
        tarfile.BLKTYPE: ResourceType.block_special_file,
        tarfile.CHRTYPE: ResourceType.character,
        tarfile.DIRTYPE: ResourceType.directory,
        tarfile.FIFOTYPE: ResourceType.fifo,
        tarfile.REGTYPE: ResourceType.file,
        tarfile.AREGTYPE: ResourceType.file,
        tarfile.SYMTYPE: ResourceType.symlink,
        tarfile.CONTTYPE: ResourceType.file,
        tarfile.LNKTYPE: ResourceType.symlink,
    }


    if six.PY2:
        def _decode(self, string):
            return string.decode(self._encoding)
    else:
        def _decode(self, string):
            return string

    def __init__(self, handle, **options):  # noqa: D102, D107
        super(TarReadFS, self).__init__(handle, **options)
        if isinstance(handle, io.IOBase):
            self._tar = TarFile.open(fileobj=handle, mode='r')
        else:
            self._tar = TarFile.open(handle, mode='r')

        self._encoding = encoding = options.get('encoding') or \
            sys.getdefaultencoding().replace('ascii', 'utf-8')

        self._members = {
            self._decode(info.name): info
                for info in self._tar.getmembers()
        }

    def exists(self, path):  # noqa: D102
        return self.isdir(path) or self.isfile(path)

    def isdir(self, path):  # noqa: D102
        _path = relpath(self.validatepath(path))
        try:
            return self._members[_path].isdir()
        except KeyError:
            return any(isbase(_path, f) for f in self._members)

    def isfile(self, path):  # noqa: D102
        _path = relpath(self.validatepath(path))
        try:
            return self._members[_path].isfile()
        except KeyError:
            return False

    def listdir(self, path):  # noqa: D102
        _path = relpath(self.validatepath(path))
        if not self.isdir(_path):
            if not self.isfile(_path):
                raise errors.ResourceNotFound(path)
            raise errors.DirectoryExpected(path)
        children = (frombase(_path, n) for n in self._members if isbase(_path, n))
        return list(unique(parts(child)[1] for child in children if relpath(child)))

    def getinfo(self, path, namespaces=None):  # noqa: D102
        namespaces = namespaces or ()
        _path = relpath(self.validatepath(path))

        if not self.exists(_path):
            raise errors.ResourceNotFound(path)

        try:
            _inferred = False
            tar_info = self._members[_path]
        except KeyError:
            _inferred = True
            tar_info = tarfile.TarInfo(_path)
            tar_info.type = tarfile.DIRTYPE

        info = {'basic': {
            'name': basename(_path),
            'is_dir': tar_info.isdir()
        }}

        if 'details' in namespaces:
            info['details'] = {
                'size': tar_info.size,
                'type': int(self._TYPE_MAP.get(
                    tar_info.type, ResourceType.unknown)),
            }
            if not _inferred:
                info['details']['modified'] = tar_info.mtime

        if 'access' in namespaces and not _inferred:
            info['access'] = {
                'gid': tar_info.gid,
                'group': tar_info.gname,
                'permissions': Permissions(mode=tar_info.mode).dump(),
                'uid': tar_info.uid,
                'user': tar_info.uname,
            }

        if 'tar' in namespaces and not _inferred:
            info['tar'] = tar_info.get_info(self._encoding) \
                          if six.PY2 else tar_info.get_info()
            info['tar'].update({
                k.replace('is', 'is_'):getattr(tar_info, k)()
                for k in dir(tar_info)
                if k.startswith('is')
            })

        return Info(info)

    def openbin(self, path, mode='r', buffering=-1, **options):  # noqa: D102
        _path = relpath(self.validatepath(path))
        _mode = Mode(mode)

        if _mode.writing:
            self._on_modification_attempt(path)
        if not self.exists(path):
            raise errors.ResourceNotFound(path)

        tar_info = self._members[_path]
        if not tar_info.isfile():
            raise errors.FileExpected(path)

        bin_file = self._tar.extractfile(tar_info)
        if six.PY2: bin_file.flush = lambda: None

        return RawWrapper(bin_file)


class TarSaver(base.ArchiveSaver):
    """A TAR archive serializer.
    """

    _compression_map = {
        '.tar': '', '.xz': 'xz', '.txz': 'xz',
        '.gz': 'gz', '.tgz':'gz', '.bz2': 'bz2', '.tbz':'bz2',
    }

    if six.PY2:
        def _encode(self, string):
            return string.encode(self.encoding)
    else:
        def _encode(self, string):
            return string

    def __init__(self, output, overwrite=False, initial_position=0, **options):  # noqa: D102, D107
        super(TarSaver, self).__init__(output, overwrite, initial_position)
        self.encoding = options.pop('encoding', 'utf-8')

        self.compression = options.pop('compression', '')

        if not self.compression and hasattr(output, 'name'):
            _, extension = splitext(output.name)
            self.compression = self._compression_map.get(extension, '')

        self.buffer_size = options.pop('buffer_size', io.DEFAULT_BUFFER_SIZE)

    def _to(self, handle, fs):  # noqa: D102
        attr_map = {
            'uid': 'uid', 'gid': 'gid', 'uname': 'user', 'gname': 'group'}
        type_map = {
            v:k for k,v in TarReadFS._TYPE_MAP.items()}

        mode = 'w:{}'.format(self.compression or '')
        if isinstance(handle, io.IOBase):
            _tar = TarFile.open(fileobj=handle, mode=mode)
        else:
            _tar = TarFile.open(handle, mode=mode)

        current_time = time.time()

        with _tar:
            for path, info in fs.walk.info(namespaces=('details', 'access', 'stat')):

                tar_info = tarfile.TarInfo(self._encode(relpath(path)))

                if info.has_namespace('stat'):
                    mtime = info.get('stat', 'st_mtime', current_time)
                else:
                    mtime = info.modified or current_time

                if isinstance(mtime, datetime.datetime):
                    mtime = datetime_to_epoch(mtime)
                if isinstance(mtime, float):
                    mtime = int(mtime)
                tar_info.mtime = mtime

                if info.has_namespace('access'):
                    for tarattr, infoattr in attr_map.items():
                        if getattr(info, infoattr) is not None:
                            setattr(tar_info, tarattr, getattr(info, infoattr))
                    tar_info.mode = getattr(info.permissions, 'mode', 0o420)


                tar_info.size = info.size
                tar_info.type = type_map.get(info.type, tarfile.REGTYPE)

                if not info.is_dir:
                    with fs.openbin(path) as bin_file:
                        _tar.addfile(tar_info, bin_file)
                else:
                    _tar.addfile(tar_info)


class TarFS(base.ArchiveFS):
    """A filesystem in a TAR archive.
    """

    _read_fs_cls = TarReadFS
    _saver_cls = TarSaver

    def __init__(self, handle, **options):  # noqa: D102, D107
        options.setdefault('compression', 'gz')
        options.setdefault('encoding', 'utf-8')
        options['proxy'] = options.pop('temp_fs', 'temp://__ziptemp__')
        super(TarFS, self).__init__(handle, **options)



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

from ...enums import Seek
from ...iotools import RawWrapper as BaseWrapper


class RawWrapper(BaseWrapper):  # noqa: D101

    def seek(self, pos, whence=Seek.set):  # noqa: D102
        if whence == Seek.set:
            if pos < 0:
                raise ValueError("Negative seek position {}".format(pos))
        elif whence == Seek.end:
            if pos > 0:
                raise ValueError("Positive seek position {}".format(pos))
        elif whence != Seek.current:
            raise ValueError(
                "Invalid whence ({}, should be {}, {} or {})".format(
                    whence, Seek.set, Seek.current, Seek.end
                )
            )
        self._f.seek(pos, whence)
        return self._f.tell()



# coding: utf-8
"""Base classes to test archive filesystems.

Adapted from the core Pyfilesystem2 test module `test_archives.py
<https://github.com/PyFilesystem/pyfilesystem2/blob/master/tests/test_archives.py>`_
written by `Will McGugan <https://github.com/willmcgugan>`_ under the **MIT**
license.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import os
import abc
import six
import sys
import shutil
import tempfile

from .. import open_fs
from .. import walk
from .. import errors
from ..test import UNICODE_TEXT
from ..enums import Seek

from . import base


@six.add_metaclass(abc.ABCMeta)
class ArchiveReadTestCases(object):
    """Base class to test ArchiveReadFS subclasses.
    """

    long_names = False
    unicode_names = False

    @abc.abstractproperty
    def _archive_read_fs(self):  # noqa: D401
        """The `ArchiveReadFS` class to test.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def compress(self, handle, fs):
        """Compress ``fs`` and write the resulting archive to ``handle``.
        """
        pass

    @abc.abstractmethod
    def remove_archive(self, handle):
        """Remove the archive in ``handle``.
        """
        pass

    @staticmethod
    def make_source_fs():
        """Create the source filesystem.
        """
        return open_fs('temp://')

    def build_source(self, fs):
        """Create the test structure in the source filesystem.
        """
        fs.makedirs('foo/bar/baz')
        fs.makedir('tmp')
        fs.settext('top.txt', 'Hello, World')
        fs.settext('top2.txt', 'Hello, World')
        fs.settext('foo/bar/egg', 'foofoo')
        fs.makedir('unicode')
        fs.makedir('extras')
        fs.settext('unicode/text.txt', UNICODE_TEXT)

        if self.unicode_names:
            fs.settext('extras/☭༉', 'I live in a unicode file !')
        if self.long_names:
            fs.settext('extras/'+'a'*300, 'This is a long file name !')

    def setUp(self, handle):  # noqa: D102
        self.handle = handle
        self.source_fs = source_fs = self.make_source_fs()
        self.build_source(source_fs)
        self.compress(self.handle, source_fs)
        self.handle.seek(0)
        self.fs = self._archive_read_fs(self.handle)

        self.assertIsInstance(
            self.fs, base.ArchiveReadFS,
            'load_archive did not return an ArchiveReadFS.'
        )

    def tearDown(self):  # noqa: D102
        self.source_fs.close()
        self.fs.close()
        self.remove_archive(self.handle)

    def test_create_failed(self):
        """Check CreateFailed is on rubbish input stream.
        """
        handle = io.BytesIO()
        handle.write(b'a'*100+b'z'*100)
        handle.seek(0)

        with self.assertRaises(errors.CreateFailed):
            self._archive_read_fs(handle)

    def test_repr(self):
        """Check that `__repr__` doesn't raise any error.
        """
        repr(self.fs)

    def test_str(self):
        """Check that `__str__` can be coerced to `six.text_type`.
        """
        self.assertIsInstance(six.text_type(self.fs), six.text_type)

    def test_scandir(self):
        """Check that `fs.base.FS.scandir` has the expected behaviour.
        """
        dirsize = self.fs.getdetails('foo').size
        if dirsize is None:
            self.skipTest("Filesystem does not support 'details' namespace")

        for entry in self.fs.scandir('/', namespaces=('details',)):
            if entry.is_dir:
                self.assertEqual(entry.size, dirsize)
            else:
                self.assertEqual(entry.size, len(self.fs.getbytes(entry.name)))

        with self.assertRaises(errors.ResourceNotFound):
            _ = next(self.fs.scandir('what'))
        with self.assertRaises(errors.DirectoryExpected):
            _ = next(self.fs.scandir('top.txt'))

    def test_readonly(self):
        """Check that read-only filesystems raise an error on edit attempts.
        """
        self.assertTrue(self.fs.getmeta().get('read_only'),
                        'Filesystem is not read-only.')

        with self.assertRaises(errors.ResourceReadOnly):
            self.fs.makedir('newdir')
        with self.assertRaises(errors.ResourceReadOnly):
            self.fs.remove('top.txt')
        with self.assertRaises(errors.ResourceReadOnly):
            self.fs.removedir('foo/bar/baz')
        with self.assertRaises(errors.ResourceReadOnly):
            self.fs.create('foo.txt')
        with self.assertRaises(errors.ResourceReadOnly):
            self.fs.setinfo('foo.txt', {})

    def test_getinfo(self):
        """Check that `fs.base.FS.getinfo` has the expected behaviour.
        """
        root = self.fs.getinfo('/')
        self.assertEqual(root.name, '')
        self.assertTrue(root.is_dir)

        top = self.fs.getinfo('top.txt', 'details')
        self.assertEqual(top.size, 12)
        self.assertFalse(top.is_dir)

        with self.assertRaises(errors.ResourceNotFound):
            _ = self.fs.getinfo('boom.txt')

    def test_listdir(self):
        """Check that `fs.base.FS.listdir` has the expected behaviour.
        """
        self.assertEqual(
            sorted(self.source_fs.listdir('/')),
            sorted(self.fs.listdir('/'))
        )
        with self.assertRaises(errors.DirectoryExpected):
            self.fs.listdir('top.txt')
        with self.assertRaises(errors.ResourceNotFound):
            self.fs.listdir('nothere')

    def test_open(self):
        """Check that `fs.base.FS.open` has the expected behaviour.
        """
        with self.fs.open('top.txt') as f:
            chars = []
            while True:
                c = f.read(2)
                if not c:
                    break
                chars.append(c)
            self.assertEqual(
                ''.join(chars),
                'Hello, World'
            )
        self.assertRaises(errors.ResourceNotFound, self.fs.open, 'nothere.txt')
        self.assertRaises(errors.FileExpected, self.fs.open, 'foo')

    def test_files_seek(self):
        """Check that handles `seek` method work as intended.
        """
        with self.fs.openbin('top.txt') as f:
            if f.seekable():
                self.assertEqual(f.seek(0), 0)
                self.assertEqual(f.tell(), 0)
                self.assertEqual(f.seek(2), 2)
                self.assertEqual(f.tell(), 2)
                self.assertEqual(f.seek(2, Seek.current), 4)
                self.assertEqual(f.tell(), 4)
                self.assertEqual(f.seek(-3, Seek.current), 1)
                self.assertEqual(f.tell(), 1)
                self.assertEqual(f.seek(-1, Seek.end), 11)
                self.assertEqual(f.tell(), 11)
                self.assertRaises(ValueError, f.seek, -3, Seek.set)
                self.assertRaises(ValueError, f.seek, 3, Seek.end)
                self.assertRaises(ValueError, f.seek, 0, 12)
            else:
                self.assertRaises(io.UnsupportedOperation, f.seek, 0)
                self.assertFalse(f.seekable())

    def test_gettext(self):
        """Check that `fs.base.FS.gettext` has the expected behaviour.
        """
        self.assertEqual(self.fs.gettext('top.txt'), 'Hello, World')
        self.assertEqual(self.fs.gettext('foo/bar/egg'), 'foofoo')
        self.assertRaises(errors.ResourceNotFound, self.fs.getbytes, 'what.txt')
        self.assertRaises(errors.FileExpected, self.fs.gettext, 'foo')

    def test_getbytes(self):
        """Check that `fs.base.FS.getbytes` has the expected behaviour.
        """
        self.assertEqual(self.fs.getbytes('top.txt'), b'Hello, World')
        self.assertEqual(self.fs.getbytes('foo/bar/egg'), b'foofoo')
        self.assertRaises(errors.ResourceNotFound, self.fs.gettext, 'what.txt')
        self.assertRaises(errors.FileExpected, self.fs.getbytes, 'foo')

    def test_walk_files(self):
        """Check that `fs.base.FS.walk.files` has the expected behaviour.
        """
        source_files = sorted(walk.walk_files(self.source_fs))
        archive_files = sorted(walk.walk_files(self.fs))

        self.assertEqual(
            source_files,
            archive_files
        )

    def test_implied_dir(self):
        """Check that implied directories are accessible from archives.
        """
        self.fs.getinfo('foo/bar')
        self.fs.getinfo('foo')

    def test_archive_meta_standard(self):
        """Check that `ArchiveReadFS.getmeta` standard *ns* has the right keys.
        """
        meta = self.fs.getmeta("standard")
        self.assertEqual(meta, self.fs.getmeta())

        self.assertIn('case_insensitive', meta)
        self.assertIn('invalid_path_chars', meta)
        self.assertIn('max_path_length', meta)
        self.assertIn('max_sys_path_length', meta)
        self.assertIn('network', meta)
        self.assertIn('read_only', meta)
        self.assertIn('supports_rename', meta)
        self.assertIn('thread_safe', meta)
        self.assertIn('unicode_paths', meta)
        self.assertIn('virtual', meta)

    def test_archive_meta_other(self):
        """Check that `getmeta` returns an empty dict on unknown namespace.
        """
        self.assertEqual(self.fs.getmeta("nonamespace"), {})

    def test_unicode_names(self):
        """Check that the filesystem accepts unicode paths.
        """
        if not self.unicode_names:
            self.skipTest("Set unicode_names to True in test case to enable")

        self.assertTrue(self.fs.exists('extras/☭༉'))
        self.assertFalse(self.fs.getinfo('extras/☭༉').is_dir)

        with self.assertRaises(errors.DirectoryExpected):
            self.fs.listdir('extras/☭༉')

        with self.fs.openbin('extras/☭༉') as f:
            self.assertEqual(f.read(), b'I live in a unicode file !')

    def test_long_names(self):
        """Check that the filesystem accepts path with more than 300 chars.
        """
        if not self.long_names:
            self.skipTest("Set long_names to True in test case to enable")

        filename = 'extras/'+'a'*300
        self.assertTrue(self.fs.exists(filename))
        self.assertFalse(self.fs.getinfo(filename).is_dir)

        with self.assertRaises(errors.DirectoryExpected):
            self.fs.listdir(filename)

        with self.fs.openbin(filename) as f:
            self.assertEqual(f.read(), b'This is a long file name !')

    def test_create_failed(self):
        """Check that the filesystem constructor raises `CreateFailed`.
        """
        self.assertRaises(errors.CreateFailed, self._archive_read_fs, None)

    def test_validatepath(self):
        """Check that validatepath works as intended.
        """
        self.assertRaises(errors.IllegalBackReference, self.fs.validatepath, '/..')
        self.assertRaises(errors.InvalidCharsInPath, self.fs.validatepath, '\0')
        self.assertEqual(self.fs.validatepath('foo/bar/../baz'), '/foo/baz')


@six.add_metaclass(abc.ABCMeta)
class ArchiveIOTestCases(object):
    """Base class to test ArchiveFS subclasses openers.
    """

    @classmethod
    def setUpClass(cls):  # noqa: D102
        cls.tempdir = tempfile.mkdtemp()

    @classmethod
    def tearDownClass(cls):  # noqa: D102
        shutil.rmtree(cls.tempdir)

    @abc.abstractproperty
    def _archive_fs(self):  # noqa: D401
        """The `ArchiveFS` class to test.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def compress(self, handle, fs):
        """Compress ``fs`` and write the resulting archive to ``handle``.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def iter_files(self, handle):
        """Return an iterator over the files in the archive in ``handle``.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def iter_dirs(self, handle):
        """Return an iterator over the directories in the archive in ``handle``.
        """
        raise NotImplementedError()

    def tearDown(self):  # noqa: D102
        if hasattr(self, 'source_fs'):
            self.source_fs.close()
        if hasattr(getattr(self, 'handle', None), 'close'):
            self.handle.close()

    def build_source(self, fs):
        """Create the test structure in the source filesystem.
        """
        fs.settext('foo.txt', 'Hello World !')
        fs.makedir('baz')
        fs.settext('baz/bar.txt', 'Savy ?')
        fs.makedir('egg')

    def make_archive(self, handle):  # noqa: D102
        self.handle = handle
        self.source_fs = source_fs = self.make_source_fs()
        self.build_source(source_fs)
        self.compress(handle, source_fs)
        if hasattr(handle, 'seek'):
            handle.seek(0)
        return handle

    def make_source_fs(self):
        """Create the source filesystem.
        """
        return open_fs('temp://')

    def _test_read(self, fs):
        self.assertEqual(set(fs.listdir('/')), {'foo.txt', 'egg', 'baz'})
        self.assertEqual(fs.gettext('foo.txt'), 'Hello World !')
        self.assertTrue(fs.exists('baz/bar.txt'))
        self.assertTrue(fs.isdir('egg'))

    def _test_read_write(self, fs):
        self._test_read(fs)
        self._test_write(fs)

    def _test_write(self, fs):
        fs.touch('ham.txt')
        fs.makedirs('/spam/qux')
        fs.touch('/spam/boom.txt')

        if fs.isfile('foo.txt'):
            fs.remove('foo.txt')
        if fs.isdir('egg'):
            fs.removedir('egg')

        self.assertTrue(fs.isdir('spam/qux'))
        self.assertTrue(fs.isfile('spam/boom.txt'))
        self.assertFalse(fs.isdir('egg') or fs.exists('egg'))
        self.assertNotIn('egg', fs.listdir('/'))

    def test_read_stream(self):
        """Check archives can be read from a stream.
        """
        stream = self.make_archive(io.BytesIO())
        with self._archive_fs(io.BufferedReader(stream), close_handle=False) as archive_fs:
            self._test_read(archive_fs)

        self.assertEqual(
            sorted(self.iter_files(stream)),
            ['/baz/bar.txt', '/foo.txt'],
        )
        self.assertEqual(
            sorted(self.iter_dirs(stream)),
            ['/baz', '/egg'],
        )

    def test_read_write_stream(self):
        """Check archives can be updated from and to a stream.
        """
        stream = self.make_archive(io.BytesIO())
        with self._archive_fs(stream, close_handle=False) as archive_fs:
            self._test_read_write(archive_fs)

        self.assertEqual(
            sorted(self.iter_files(stream)),
            ['/baz/bar.txt', '/ham.txt', '/spam/boom.txt'],
        )
        self.assertEqual(
            sorted(self.iter_dirs(stream)),
            ['/baz', '/spam', '/spam/qux'],
        )

    def test_write_stream(self):
        """Check archives can be written to a stream.
        """
        stream = io.BytesIO()
        stream.readable = lambda: False   # mock a write-only stream
        with self._archive_fs(stream, close_handle=False) as zipfs:
            self._test_write(zipfs)

        self.assertEqual(
            sorted(self.iter_files(stream)),
            ['/ham.txt', '/spam/boom.txt'],
        )
        self.assertEqual(
            sorted(self.iter_dirs(stream)),
            ['/spam', '/spam/qux'],
        )

    def test_read_file(self):
        """Check archives can be read from a file.
        """
        filename = os.path.join(self.tempdir, 'read')
        with self._archive_fs(self.make_archive(filename)) as zipfs:
            self._test_read(zipfs)

        self.assertEqual(
            sorted(self.iter_files(filename)),
            ['/baz/bar.txt', '/foo.txt'],
        )
        self.assertEqual(
            sorted(self.iter_dirs(filename)),
            ['/baz', '/egg'],
        )

    def test_read_write_file(self):
        """Check archives can be updated from and to a stream.
        """
        filename = os.path.join(self.tempdir, 'write')
        with self._archive_fs(self.make_archive(filename)) as zipfs:
            self._test_read_write(zipfs)

        self.assertEqual(
            sorted(self.iter_files(filename)),
            ['/baz/bar.txt', '/ham.txt', '/spam/boom.txt'],
        )
        self.assertEqual(
            sorted(self.iter_dirs(filename)),
            ['/baz', '/spam', '/spam/qux'],
        )

    def test_write_file(self):
        """Check archives can be writtent to a stream.
        """
        filename = os.path.join(self.tempdir, 'read_write')
        with self._archive_fs(filename) as zipfs:
            self._test_write(zipfs)
            print(zipfs._saver)

        self.assertEqual(
            sorted(self.iter_dirs(filename)),
            ['/spam', '/spam/qux']
        )
        self.assertEqual(
            sorted(self.iter_files(filename)),
            ['/ham.txt', '/spam/boom.txt']
        )

    def test_iter_dirs(self):
        """Check the `iter_dirs` method works as intented.
        """
        handle = self.make_archive(io.BytesIO())

        self.assertEqual(
            sorted(self.iter_dirs(handle)),
            sorted(self.source_fs.walk.dirs()),
        )

    def test_iter_files(self):
        """Check the `iter_files` method works as intended.
        """
        handle = self.make_archive(io.BytesIO())

        self.assertEqual(
            sorted(self.iter_files(handle)),
            sorted(self.source_fs.walk.files())
        )



# coding: utf-8
"""ISO Disk Image filesystems.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import os
import re
import operator
import weakref

import six
import pycdlib

from ... import errors
from ...mode import Mode
from ...info import Info
from ...path import recursepath, iteratepath, join, frombase
from ...enums import ResourceType, Seek

from .. import base

from ._utils import iso_path_slugify


class ISOFile(io.RawIOBase):
    """A read-only, seekable file on an ISO filesystem.
    """

    def __init__(self, fs, entry):  # noqa: D102, D107

        self._fs = fs
        self._cd = cd = fs._cd
        self._entry = entry

        self._handle, self._size = entry.data_fp, entry.data_length

        if entry.original_data_location == entry.DATA_ON_ORIGINAL_ISO:
            self._start = entry.orig_extent_loc * cd.pvd.logical_block_size()
        else:
            self._start = entry.fp_offset
        self._position = 0
        self._end = self._start + self._size

    def readable(self):  # noqa: D102
        return True

    def writable(self):  # noqa: D102
        return False

    def read(self, size=-1):  # noqa: D102
        with self._fs.lock():
            self._handle.seek(self._start + self._position)
            if size == -1 or self._position + size > self._size:
                size = self._size - self._position
            self._position += size
            return self._handle.read(size)

    def seek(self, offset, whence=Seek.set):  # noqa: D102
        if whence == Seek.set:
            if offset < 0:
                raise ValueError("Negative seek position {}".format(offset))
            self._position = min(offset, self._size)
        elif whence == Seek.current:
            self._position = max(min(self._position + offset, self._size), 0)
        elif whence == Seek.end:
            if offset > 0:
                raise ValueError("Positive seek position {}".format(offset))
            self._position = max(0, self._size + offset)
        else:
            raise ValueError(
                "Invalid whence ({}, should be {}, {} or {})".format(
                    whence, Seek.set, Seek.current, Seek.end
                )
            )
        return self._position

    def seekable(self):  # noqa: D102
        return True

    def tell(self):  # noqa: D102
        return self._position


class ISOReadFS(base.ArchiveReadFS):
    """A read-only ISO filesystem.
    """

    _meta = {
        'standard': {
            'case_insensitive': True,
            'network': False,
            'read_only': True,
            'supports_rename': False,
            'thread_safe': True,
            'unicode_paths': True,
            'virtual': False,
            'max_path_length': None,
            'max_sys_path_length': None,
            'invalid_path_chars': '\0',
        },
    }

    def _get_name_from_entry(self, entry):
        if self._rock_ridge:
            return entry.rock_ridge.name().decode('utf-8')
        elif self._joliet:
            return entry.file_identifier().decode('utf-16be')
        else:
            return entry.file_identifier().decode('ascii')\
                        .lower().rstrip(';1').rstrip('.')

    def _get_cd_entry(self, path):

        # Get the closest parent of the requested path
        for subpath in recursepath(path):
            if subpath in self._path_table:
                entry = self._path_table[subpath]
                _path = frombase(subpath, path)
                break

        # If the actual entry is found, return it directly
        if not _path:
            return entry

        # Else, recurse down from the closest transitive parent entry
        for name in iteratepath(_path):

            # Get the content of the CWD and store each entry in the path table
            for child in entry.children:
                if not child.is_dot() and not child.is_dotdot():
                    child_name = self._get_name_from_entry(child)
                    child_path = join(subpath, child_name)
                    self._path_table[child_path] = child

            # Raise an error if no entry is found with the given name
            entry = self._path_table.get(join(subpath, name), None)
            if entry is None:
                raise errors.ResourceNotFound(path)

            # Raise an error if the non-final entry is not a directory
            if path != join(subpath, name) and not entry.is_dir():
                raise errors.DirectoryExpected(join(subpath, name))

            # Move one level deeper
            subpath = join(subpath, name)

        return entry

    def _get_info_from_entry(self, entry, namespaces=None):
        namespaces = namespaces or ()

        info = {'basic': {
            'name': self._get_name_from_entry(entry),
            'is_dir': entry.is_dir()
        }}

        # TODO: the rest
        if 'details' in namespaces:
            info['details'] = details = {
                'size': entry.data_length,
                'type': ResourceType.directory if entry.is_dir() else ResourceType.file,
            }

            if self._rock_ridge and entry.rock_ridge.is_symlink():
                details['type'] = ResourceType.symlink


        return Info(info)

    def __init__(self, handle, **options):  # noqa: D102, D107
        super(ISOReadFS, self).__init__(handle, **options)
        self._cd = pycdlib.PyCdlib()

        if isinstance(handle, io.IOBase):
            self._cd.open_fp(handle)
        else:
            self._cd.open(handle)

        self._joliet = self._cd.joliet_vd is not None
        self._rock_ridge = self._cd.rock_ridge is not None
        self._joliet_only = self._joliet and not self._rock_ridge

        self._path_table = weakref.WeakValueDictionary()
        self._path_table['/'] = self._cd.get_entry('/', self._joliet_only)

    def getinfo(self, path, namespaces=None):  # noqa: D102
        _path = self.validatepath(path)

        if _path in '/':
            return Info({
                'basic': {'name': '', 'is_dir': True},
                'details': {'size': 0, 'type': ResourceType.directory}
            })
        else:
            entry = self._get_cd_entry(_path)
            return self._get_info_from_entry(entry, namespaces)

    def scandir(self, path, namespaces=None, page=None):  # noqa: D102
        _path = self.validatepath(path)

        entry = self._get_cd_entry(_path)

        if entry.is_file():
            raise errors.DirectoryExpected(path)

        for child in entry.children:
            if not child.is_dot() and not child.is_dotdot():
                yield self._get_info_from_entry(child, namespaces)

    def listdir(self, path):  # noqa: D102
        return [child.name for child in self.scandir(path)]

    def openbin(self, path, mode='r', buffering=-1, **options):  # noqa: D102
        _path = self.validatepath(path)
        _mode = Mode(mode)

        if _mode.writing:
            self._on_modification_attempt(path)

        if self.isdir(_path):
            raise errors.FileExpected(path)
        elif not self.isfile(_path):
            raise errors.ResourceNotFound(path)

        print("IN OPENBIN: ", _path)

        entry = self._get_cd_entry(_path)
        return ISOFile(self, entry)

    def getmeta(self, namespace="standard"):  # noqa: D102
        meta = self._meta.get(namespace, {}).copy()
        if namespace == "standard" and not (self._rock_ridge or self._joliet):
            meta['case_insensitive'] = self._cd.interchange_level < 4
            meta['max_path_length'] = 255
        return meta

    def getsize(self, path):  # noqa: D102
        _path = self.validatepath(path)
        entry = self._get_cd_entry(_path)
        return entry.file_length()


class ISOSaver(base.ArchiveSaver):
    """An ISO-9660 serializer.
    """

    def __init__(self, output, overwrite=False, initial_position=0, **options):  # noqa: D102, D107
        super(ISOSaver, self).__init__(output, overwrite, initial_position)

        self.joliet = options.pop('joliet', False)
        self.rock_ridge = options.pop('rock_ridge', '1.12')
        self.interchange_level = options.pop('interchange_level', 1)
        self.strict = self.interchange_level < 4

    def _to(self, handle, fs):
        _cd = pycdlib.PyCdlib()
        _cd.new(
            interchange_level=self.interchange_level,
            sys_ident='',
            vol_ident='',
            set_size=1,
            seqnum=1,
            log_block_size=2048,
            vol_set_ident=' ',
            pub_ident_str='',
            preparer_ident_str='',
            app_ident_str='PyCdlib (C) 2015-2016 Chris Lalancette',
            copyright_file='',
            abstract_file='',
            bibli_file='',
            vol_expire_date=None,
            app_use='',
            joliet=self.joliet,
            rock_ridge=self.rock_ridge,
            xa=False
        )
        slug_table = {'/': '/'}

        try:
            for parent, dirs, files in fs.walk('/',
                    search='breadth', namespaces=('details', 'access', 'stat')):
                for d in dirs:
                    path = join(parent, d.name)
                    iso_path = \
                        iso_path_slugify(path, slug_table, True, self.strict)
                    _cd.add_directory(
                        iso_path=iso_path,
                        rr_name=d.name if self.rock_ridge else None,
                        joliet_path=path if self.joliet else None,
                    )
                for f in files:
                    path = join(parent, f.name)
                    iso_path = \
                        iso_path_slugify(path, slug_table, strict=self.strict)
                    _cd.add_fp(
                        fp=fs.openbin(path), length=f.size, iso_path=iso_path,
                        rr_name=f.name if self.rock_ridge else None,
                        joliet_path=path if self.joliet else None,
                    )
        finally:
            if isinstance(handle, io.IOBase):
                _cd.write_fp(handle)
            else:
                _cd.write(handle)
            _cd.close()


class ISOFS(base.ArchiveFS):
    """An ISO-9660 filesystem.
    """

    _read_fs_cls = ISOReadFS
    _saver_cls = ISOSaver



# coding: utf-8
"""Declaration of the `open_archive` function.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import six
import pkg_resources

from .. import errors
from ..path import basename
from ..opener import open_fs
from ..opener.errors import UnsupportedProtocol

from . import base


def open_archive(fs_url, archive):
    """Open an archive on a filesystem.

    This function tries to mimick the behaviour of `fs.open_fs` as closely
    as possible: it accepts either a FS URL or a filesystem instance, and
    will close all resources it had to open.

    Arguments:
        fs_url (FS or text_type): a FS URL, or a filesystem
            instance, where the archive file is located.
        archive (text_type): the path to the archive file on the
            given filesystem.

    Raises:
        `fs.opener._errors.Unsupported`: when the archive type is not supported
            (either the file extension is unknown or the opener requires unmet
            dependencies).

    Example:
        >>> from fs.archive import open_archive
        >>> with open_archive('mem://', 'test.tar.gz') as archive_fs:
        ...     type(archive_fs)
        <class 'fs.archive.tarfs.TarFS'>

    Hint:
        This function finds the entry points defined in group
        ``fs.archive.open_archive``, using the names of the entry point
        as the registered extension.

    """
    it = pkg_resources.iter_entry_points('fs.archive.open_archive')
    entry_point = next((ep for ep in it if archive.endswith(ep.name)), None)

    if entry_point is None:
        raise UnsupportedProtocol(
            'unknown archive extension: {}'.format(archive))

    try:
        archive_opener = entry_point.load()
    except pkg_resources.DistributionNotFound as df: # pragma: no cover
        six.raise_from(UnsupportedProtocol(
            'extension {} requires {}'.format(entry_point.name, df.req)), None)

    try:
        binfile = None
        archive_fs = None
        fs = open_fs(fs_url)

        if issubclass(archive_opener, base.ArchiveFS):
            try:
                binfile = fs.openbin(archive, 'r+')
            except errors.ResourceNotFound:
                binfile = fs.openbin(archive, 'w')
            except errors.ResourceReadOnly:
                binfile = fs.openbin(archive, 'r')
                archive_opener = archive_opener._read_fs_cls

        elif issubclass(archive_opener, base.ArchiveReadFS):
            binfile = fs.openbin(archive, 'r')

        if not hasattr(binfile, 'name'):
            binfile.name = basename(archive)

        archive_fs = archive_opener(binfile)

    except Exception:
        getattr(archive_fs, 'close', lambda: None)()
        getattr(binfile, 'close', lambda: None)()
        raise

    else:
        return archive_fs



# coding: utf-8
"""Zip archive filesystems.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import sys
import six
import time
import shutil
import zipfile
import datetime

from ... import errors
from ...info import Info
from ...mode import Mode
from ...time import datetime_to_epoch
from ...path import forcedir, relpath, dirname, basename, abspath
from ...path import iteratepath, recursepath, frombase, join
from ...enums import ResourceType
from ...iotools import RawWrapper
from ..._fscompat import fsdecode, fsencode

from .. import base


class ZipReadFS(base.ArchiveReadFS):
    """A read-only filesystem within a ZIP archive.
    """

    _meta = {
        'standard': {
            'case_insensitive': True,
            'network': False,
            'read_only': True,
            'supports_rename': False,
            'thread_safe': True,
            'unicode_paths': True,
            'virtual': False,
            'max_path_length': None,
            'max_sys_path_length': None,
            'invalid_path_chars': '\x00\x01',
        },
    }

    def __init__(self, handle, **options):  # noqa: D102, D107
        super(ZipReadFS, self).__init__(handle, **options)
        self._zip = zipfile.ZipFile(self._handle)

        self._encoding = encoding = options.get('encoding') or \
            sys.getdefaultencoding().replace('ascii', 'utf-8')

        self._namelist = self._get_namelist(self._encoding)
        self._contents = self._get_contents(self._namelist)

    def _get_contents(self, namelist):
        contents = set()
        for name in self._namelist:
            for partial_name in recursepath(name):
                partial_name = relpath(partial_name)
                if partial_name != name:
                    partial_name = forcedir(partial_name)
                if partial_name:
                    contents.add(partial_name)
        return contents

    def _get_namelist(self, encoding):
        if six.PY2:
            return [n.decode(encoding) for n in self._zip.namelist()]
        else:
            return list(self._zip.namelist())

    def getinfo(self, path, namespaces=None):  # noqa: D102
        namespaces = namespaces or ()
        _path = self.validatepath(path)

        if not self.exists(_path):
            raise errors.ResourceNotFound(path)

        info = {'basic': {
            'name': basename(_path),
            'is_dir': self.isdir(_path),
        }}

        if namespaces:

            try:
                zip_name = relpath(
                    forcedir(_path) if self.isdir(_path) else _path)
                zip_info = self._zip.getinfo(zip_name)
            except KeyError:
                if info['basic']['is_dir']: # Implicit directory
                    info['details'] = {
                        'size': 0,
                        'type': ResourceType.directory,
                    }
            else:
                modified_epoch = datetime_to_epoch(
                    datetime.datetime(*zip_info.date_time)
                )
                info['zip'] = {
                    k: getattr(zip_info, k)
                    for k in dir(zip_info)
                    if (not k.startswith('_') and
                        not callable(getattr(zip_info, k)))
                }
                info['details'] = {
                    'size': zip_info.file_size,
                    'type': int(
                        ResourceType.directory
                        if zip_info.filename.endswith('/') else
                        ResourceType.file
                    ),
                    'modified': modified_epoch
                }

        return Info(info)

    def getbytes(self, path):  # noqa: D102
        _path = self.validatepath(path)

        if self.isdir(_path):
            raise errors.FileExpected(path)
        if not self.isfile(_path):
            raise errors.ResourceNotFound(path)

        zip_bytes = self._zip.read(relpath(_path))
        return zip_bytes

    def isfile(self, path):  # noqa: D102
        _path = self.validatepath(path).lower()
        return relpath(_path) in self._contents

    def isdir(self, path):  # noqa: D102
        if path in '/':
            return True
        _path = self.validatepath(path).lower()
        return relpath(forcedir(_path)) in self._contents

    def exists(self, path):  # noqa: D102
        if path in '/':
            return True
        return self.isdir(path) or self.isfile(path)

    def listdir(self, path):  # noqa: D102
        return [entry.name for entry in self.scandir(path)]

    def scandir(self, path, namespaces=None, page=None):  # noqa: D102
        _path = self.validatepath(path)

        if not self.exists(path):
            raise errors.ResourceNotFound(path)
        elif not self.isdir(path):
            raise errors.DirectoryExpected(path)

        seen = set()

        for fullname in self._namelist:
            fullname = abspath(fullname.rstrip('/'))

            if fullname.startswith(_path) and fullname != _path:

                name = iteratepath(relpath(frombase(_path, fullname)))[0]
                fullname = join(_path, name)

                if namespaces is None:
                    info = Info({'basic': {
                        'name': name,
                        'is_dir': self.isdir(fullname)
                    }})
                else:
                    info = self.getinfo(fullname, namespaces=namespaces)

                if fullname not in seen:
                    seen.add(fullname)
                    yield info

    def openbin(self, path, mode='r', buffering=-1, **options):  # noqa: D102
        _path = relpath(self.validatepath(path))
        _mode = Mode(mode)

        if _mode.writing:
            self._on_modification_attempt(path)

        if self.isdir(_path):
            raise errors.FileExpected(path)
        if not self.isfile(_path):
            raise errors.ResourceNotFound(path)

        if six.PY2:
            _path = _path.encode(self._encoding)

        bin_file = self._zip.open(_path, 'r')
        return RawWrapper(bin_file)

    def close(self):  # noqa: D102
        if not self.isclosed():
            super(ZipReadFS, self).close()
            self._zip.close()


class ZipSaver(base.ArchiveSaver):
    """A ZIP archive serializer.
    """

    def __init__(self, output, overwrite=False, initial_position=0, **options):  # noqa: D102, D107
        super(ZipSaver, self).__init__(output, overwrite, initial_position)
        self.encoding = options.pop('encoding', 'utf-8')
        self.compression = options.pop('compression', zipfile.ZIP_DEFLATED)
        self.buffer_size = options.pop('buffer_size', io.DEFAULT_BUFFER_SIZE)

    def _to(self, handle, fs):  # noqa: D102
        _zip = zipfile.ZipFile(
            handle, mode='w', compression=self.compression, allowZip64=True)

        with _zip:

            for path, info in fs.walk.info(namespaces=('details', 'stat')):

                # Zip names must be relative, directory names must end
                # with a slash.
                zip_name = relpath(forcedir(path) if info.is_dir else path)
                if six.PY2:
                    # Python2 expects bytes filenames
                    zip_name = zip_name.encode(self.encoding, 'replace')

                if info.has_namespace('stat'):
                    # get zip time directory from the stat structure
                    st_mtime = info.get('stat', 'st_mtime', None)
                    _mtime = time.localtime(st_mtime)
                    zip_time = _mtime[0:6]

                else:
                    # use the modified time from details namespace.
                    mt = info.modified or datetime.datetime.utcnow()
                    zip_time = (
                        mt.year, mt.month, mt.day,
                        mt.hour, mt.minute, mt.second
                    )

                # create a custom zip_info
                zip_info = zipfile.ZipInfo(zip_name, zip_time)

                if info.is_dir:
                    # only write empty directories (other are implicit)
                    if next(fs.walk.files(path), None) is None:
                        _zip.writestr(zip_info, b'')
                else:
                    #
                    size = fs.getsize(path)
                    if size is not None and size > 0:
                        zip_info.file_size = size


                    with fs.openbin(path, 'rb') as src_file:
                        self._write_to_zip(_zip, zip_info, src_file)
                        # with _zip.open(zip_info, 'w') as dst_file:
                        #     shutil.copyfileobj(src_file, dst_file, self.buffer_size)

    if sys.version_info >= (3, 6):
        def _write_to_zip(self, _zip, zip_info, src_file):
            with _zip.open(zip_info, 'w') as dst_file:
                shutil.copyfileobj(src_file, dst_file, self.buffer_size)
    else:
        def _write_to_zip(self, _zip, zip_info, src_file):
            _zip.writestr(zip_info, src_file.read())


class ZipFS(base.ArchiveFS):
    """A filesystem within a ZIP archive.
    """

    _read_fs_cls = ZipReadFS
    _saver_cls = ZipSaver

    def __init__(self, handle, **options):  # noqa: D102, D107
        options.setdefault('compression', zipfile.ZIP_DEFLATED)
        options.setdefault('encoding', 'utf-8')
        options['proxy'] = options.pop('temp_fs', 'temp://__ziptemp__')
        super(ZipFS, self).__init__(handle, **options)



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import operator
import itertools
import abc
import six

from .. import errors
from ..base import FS
from ..errors import ResourceNotFound
from ..copy import copy_file
from ..path import abspath, dirname, join, normpath
from ..mode import Mode
from ..opener import open_fs
from ..wrapfs import WrapFS

from ._utils import unique, UniversalContainer, NoWrapMeta


__all__ = ["WrapWritable"]




@six.add_metaclass(NoWrapMeta)
class WrapWritable(WrapFS):
    """A wrapper that makes a read-only FS writable.

    All modifications are discarded once the filesystem is closed. Using
    the `NoWrapMeta` metaclass, the wrapper will use the base `FS`
    implementation of the non-essential methods instead of using the
    `WrapFS` implementation.
    """

    def __init__(self, delegate_fs, writable_fs="mem://"):  # noqa: D107
        super(WrapWritable, self).__init__(delegate_fs)
        self._rfs = delegate_fs
        self._wfs = open_fs(writable_fs)
        self._removed = set()

    def appendbytes(self, path, data):  # noqa: D102
        _path = self.validatepath(path)
        if not isinstance(data, six.binary_type):
            raise TypeError("must be bytes")
        if not self.isdir(dirname(_path)):
            raise errors.ResourceNotFound(dirname(path))
        self._wfs.makedirs(dirname(_path), recreate=True)
        if self.exists(_path) and not self.isfile(_path):
            raise errors.FileExpected(path)
        if self._rfs.isfile(_path) and _path not in self._removed:
            if not self._wfs.isfile(_path):
                _copy_file_rich(self._rfs, _path, self._wfs)
        if _path in self._removed:
            self._removed.remove(_path)
        return self._wfs.appendbytes(_path, data)

    # def appendtext(self, path, text):
    #     _path = self.validatepath(path)
    #     if not isinstance(text, six.text_type):
    #         raise TypeError("must be unicode string")
    #     if not self.isdir(dirname(_path)):
    #         raise errors.ResourceNotFound(dirname(path))
    #     self._wfs.makedirs(dirname(_path), recreate=True)
    #     if self.exists(_path) and not self.isfile(_path):
    #         raise errors.FileExpected(path)
    #     if self._rfs.isfile(_path) and _path not in self._removed:
    #         if not self._wfs.isfile(_path):
    #             _copy_file_rich(self._rfs, _path, self._wfs)
    #     if _path in self._removed:
    #         self._removed.remove(_path)
    #     return self._wfs.appendtext(_path, text)

    def close(self):  # noqa: D102
        if not self.isclosed():
            self._wfs.close()
            super(WrapWritable, self).close()

    def exists(self, path):  # noqa: D102
        _path = self.validatepath(path)
        if self._wfs.exists(path):
            return True
        if self._rfs.exists(path) and _path not in self._removed:
            return True
        return False

    def getinfo(self, path, namespaces=None):  # noqa: D102
        _path = self.validatepath(path)
        if not self.exists(_path):
            raise errors.ResourceNotFound(path)
        if self._wfs.exists(_path):
            return self._wfs.getinfo(_path, namespaces)
        return self._rfs.getinfo(_path, namespaces)

    def listdir(self, path):  # noqa: D102
        _path = self.validatepath(path)

        if not self.getinfo(path).is_dir:
            raise errors.DirectoryExpected(path)

        files_it = itertools.chain()
        if self._wfs.isdir(_path):
            files_it = itertools.chain(files_it, self._wfs.listdir(_path))
        if self._rfs.isdir(_path) and _path not in self._removed:
            files_it = itertools.chain(files_it, self._rfs.listdir(_path))

        def was_removed(name):
            return join(_path, name) in self._removed
        return list(unique(six.moves.filterfalse(was_removed, files_it)))

    def makedir(self, path, permissions=None, recreate=False):  # noqa: D102

        _path = self.validatepath(path)

        if self.exists(_path):
            if not recreate:
                raise errors.DirectoryExists(path)
        elif not self.exists(dirname(_path)):
            raise errors.ResourceNotFound(dirname(path))
        elif self.isfile(dirname(_path)):
            raise errors.DirectoryExpected(dirname(path))

        if _path in self._removed:
            self._removed.remove(_path)

        # FIXME: possible permission mismatch
        self._wfs.makedirs(dirname(_path), recreate=True)
        return self._wfs.makedir(_path, permissions, recreate)

    def openbin(self, path, mode='r', buffering=-1, **options):  # noqa: D102
        _path = self.validatepath(path)
        _mode = Mode(mode)
        _mode.validate_bin()

        if not self.exists(_path):
            if not self.isdir(dirname(_path)):
                raise ResourceNotFound(dirname(path))
            if _mode.create:
                if _path in self._removed:
                    self._removed.remove(_path)
                self._wfs.makedirs(dirname(_path), recreate=True)
                return self._wfs.openbin(path, mode, buffering, **options)
            else:
                raise ResourceNotFound(path)
        elif self._wfs.exists(_path):
            return self._wfs.openbin(path, mode, buffering, **options)
        elif not _mode.writing:
            return self._rfs.openbin(path, mode, buffering, **options)
        else:
            self._wfs.makedirs(dirname(_path), recreate=True)
            _copy_file_rich(self._rfs, _path, self._wfs)
            return self._wfs.openbin(path, mode, buffering, **options)

    def remove(self, path):  # noqa: D102
        _path = self.validatepath(path)
        if not self.getinfo(path).is_file:
            raise errors.FileExpected(path)
        self._removed.add(_path)
        if self._wfs.isfile(_path):
            self._wfs.remove(_path)

    def removedir(self, path):  # noqa: D102
        _path = self.validatepath(path)
        if not self.isempty(_path):
            raise errors.DirectoryNotEmpty(path)
        self._removed.add(_path)
        if self._wfs.isdir(_path):
            self._wfs.removedir(_path)

    def scandir(self, path, namespaces=None, page=None):  # noqa: D102
        _path = self.validatepath(path)

        if not self.exists(_path):
            raise errors.ResourceNotFound(path)
        if not self.isdir(_path):
            raise errors.DirectoryExpected(path)

        it = itertools.chain()
        if self._wfs.isdir(_path):
            it = itertools.chain(it, self._wfs.scandir(_path, namespaces))
        if self._rfs.isdir(_path) and _path not in self._removed:
            it = itertools.chain(it, self._rfs.scandir(_path, namespaces))

        def _exists(info):
            return self.exists(join(_path, info.name))
        it = six.moves.filter(_exists, it)
        it = unique(it, key=operator.attrgetter('name'))
        if page is not None:
            it = itertools.islice(it, page[0], page[1])

        return it

    def setinfo(self, path, info):  # noqa: D102
        _path = self.validatepath(path)
        if not self.exists(_path):
            raise errors.ResourceNotFound(path)
        if self._rfs.exists(_path):
            self._wfs.makedirs(dirname(_path), recreate=True)
            _copy_file_rich(self._rfs, _path, self._wfs, _path)
        return self._wfs.setinfo(_path, info)

    # def touch(self, path):
    #     _path = self.validatepath(path)
    #     if not self.isfile(_path):
    #         with self.openbin(_path, 'wb') as f:
    #             f.write(b'')

    def validatepath(self, path):  # noqa: D102
        super(WrapWritable, self).validatepath(path)
        return abspath(normpath(path))



def _copy_file_rich(src_fs, src_path, dst_fs, dst_path=None):
    dst_path = src_path if dst_path is None else dst_path
    copy_file(src_fs, src_path, dst_fs, dst_path)
    src_info = src_fs.getinfo(src_path, namespaces=UniversalContainer())
    dst_fs.setinfo(dst_path, src_info.raw)



# coding: utf-8
"""`tarfile.TarFile` subclass overloaded with additional compression openers.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import six
import tarfile

from .._utils import import_from_names

lzma = import_from_names('lzma', 'backports.lzma')


class TarFile(tarfile.TarFile):  # noqa: D101

    OPEN_METH = {
        "tar": "taropen",
        "gz": "gzopen",
        "bz2": "bz2open",
    }

    if six.PY3:
        OPEN_METH["xz"] = "xzopen"

    if lzma is not None:
        OPEN_METH["xz"] = "xzopen"

        @classmethod
        def xzopen(cls, name, mode="r", fileobj=None, preset=None, **kwargs):
            """Open lzma compressed tar archive name for reading or writing.

            Attention:
                Appending is not allowed.

            Note:
               Backported from `Python 3.6
               <https://github.com/python/cpython/blob/3.6/Lib/tarfile.py>`_
            """
            if mode not in ("r", "w", "x"):
                raise ValueError("mode must be 'r', 'w' or 'x'")

            fileobj = lzma.LZMAFile(fileobj or name, mode, preset=preset)

            try:
                t = cls.taropen(name, mode, fileobj, **kwargs)
            except (lzma.LZMAError, EOFError):
                fileobj.close()
                if mode == 'r':
                    raise tarfile.ReadError("not an lzma file")
                raise
            except:
                fileobj.close()
                raise
            t._extfileobj = False
            return t



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import string

from ...path import split, join



def iso_name_slugify(name):
    """Slugify a name in the ISO-9660 way.

    Example:
        >>> slugify('épatant')
        "_patant"
    """
    name = name.encode('ascii', 'replace').replace(b'?', b'_')
    return name.decode('ascii')


def iso_name_increment(name, is_dir=False, max_length=8):
    """Increment an ISO name to avoid name collision.

    Example:
        >>> iso_name_increment('foo.txt')
        'foo1.txt'
        >>> iso_name_increment('bar10')
        'bar11'
        >>> iso_name_increment('bar99', max_length=5)
        'ba100'
    """
    # Split the extension if needed
    if not is_dir and '.' in name:
        name, ext = name.rsplit('.')
        ext = '.{}'.format(ext)
    else:
        ext = ''

    # Find the position of the last letter
    for position, char in reversed(list(enumerate(name))):
        if char not in string.digits:
            break

    # Extract the numbers and the text from the name
    base, tag = name[:position+1], name[position+1:]
    tag = str(int(tag or 0) + 1)

    # Crop the text if the numbers are too long
    if len(tag) + len(base) > max_length:
        base = base[:max_length - len(tag)]

    # Return the name with the extension
    return ''.join([base, tag, ext])


def iso_path_slugify(path, path_table, is_dir=False, strict=True):
    """Slugify a path, maintaining a map with the previously slugified paths.

    The path table is used to prevent slugified names from collisioning,
    using the `iso_name_increment` function to deduplicate slugs.

    Example:
        >>> path_table = {'/': '/'}
        >>> iso_path_slugify('/ébc.txt', path_table)
        '/_BC.TXT'
        >>> iso_path_slugify('/àbc.txt', path_table)
        '/_BC2.TXT'
    """
    # Split the path to extract the parent and basename
    parent, base = split(path)

    # Get the parent in slugified form
    slug_parent = path_table[parent]

    # Slugify the base name
    if is_dir:
        slug_base = iso_name_slugify(base)[:8]
    else:
        name, ext = base.rsplit('.', 1) if '.' in base else (base, '')
        slug_base = '.'.join([iso_name_slugify(name)[:8], ext])
    if strict:
        slug_base = slug_base.upper()

    # Deduplicate slug if needed and update path_table
    slugs = set(path_table.values())
    path_table[path] = slug = join(slug_parent, slug_base)
    while slug in slugs:
        slug_base = iso_name_increment(slug_base, is_dir)
        path_table[path] = slug = join(slug_parent, slug_base)

    # Return the unique slug
    return slug



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import abc
import collections
import importlib
import io
import os
import sys
import typing

from six.moves import filterfalse

from ..base import FS
from ..wrapfs import WrapFS



__all__ = [
    'UniversalContainer',
    'NoWrapMeta',
    'unique',
    'import_from_names',
    'writable_path',
    'writable_stream'
]


class UniversalContainer(collections.Container):
    """A container that contains everything.

    Example:
        >>> c = UniversalContainer()
        >>> 1 in c
        True
        >>> None in c
        True
        >>> c in c
        True

    """

    def __contains__(self, _):  # noqa: D105
        return True


class NoWrapMeta(typing.GenericMeta):
    """Prevent classes from using `WrapFS` implementations of the `FS` methods.
    """

    def __new__(mcs, name, bases, attrs): # noqa: D102,D105
        _bases = bases + (FS,)
        exclude = ('__init__',)
        for base in _bases:
            if base is not WrapFS:
                for k,v in vars(base).items():
                    if callable(v) and k not in exclude:
                        attrs.setdefault(k, v)
        return super(NoWrapMeta, mcs).__new__(mcs, name, bases, attrs)


def unique(iterable, key=None):
        """Yield unique elements, preserving order.
        """
        seen = set()
        seen_add = seen.add
        if key is None:
            for element in filterfalse(seen.__contains__, iterable):
                seen_add(element)
                yield element
        else:
            for element in iterable:
                k = key(element)
                if k not in seen:
                    seen_add(k)
                    yield element


def import_from_names(*names):
    """Try to import the same function from various names.

    Example:
        >>> etree = import_from_names(
        ...     'lxml.etree',
        ...     'xml.etree.cElementTree',
        ...     'xml.etree.ElementTree'
        ... )
    """
    for name in names:
        try:
            return importlib.import_module(name)
        except ImportError:
            continue
    return None


def writable_path(path):
    """Test whether a path can be written to.
    """
    if os.path.exists(path):
        return os.access(path, os.W_OK)
    try:
        with open(path, 'w'):
            pass
    except (OSError, IOError):
        return False
    else:
        os.remove(path)
        return True


def writable_stream(handle):
    """Test whether a stream can be written to.
    """
    if isinstance(handle, io.IOBase) and sys.version_info >= (3, 5):
        return handle.writable()
    try:
        handle.write(b'')
    except (io.UnsupportedOperation, IOError):
        return False
    else:
        return True
